import unittest
from unittest.mock import patch

from fault_injection import (
    delay,
    delay_inline,
    delay_random,
    delay_random_inline,
    delay_random_norm,
    delay_random_norm_inline,
    delay_at_nth_call,
    delay_at_nth_call_inline
)


class TestDelayDecorator(unittest.TestCase):
    def test_delay_rejects_negative_time(self):
        with self.assertRaisesRegex(ValueError, "delay should have positive time_s"):
            delay(-0.1)

    def test_delay_calls_sleep_and_returns_wrapped_value(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            @delay(0.25)
            def add(a, b):
                return a + b

            self.assertEqual(add(1, 3), 4)
            sleep_mock.assert_called_once_with(0.25)

    def test_delay_disable_skips_sleep(self):
        with patch(
            "fault_injection.delays.time.sleep",
            side_effect=AssertionError("time.sleep should not be called when disabled"),
        ):
            @delay(0.25, disable=True)
            def add(a, b):
                return a + b

            self.assertEqual(add(2, 5), 7)


class TestDelayInline(unittest.TestCase):
    def test_delay_inline_rejects_negative_time(self):
        with self.assertRaisesRegex(ValueError, "delay should have positive time_s"):
            delay_inline(-0.1)

    def test_delay_inline_calls_sleep(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            self.assertIsNone(delay_inline(0.25))
            sleep_mock.assert_called_once_with(0.25)

    def test_delay_inline_disable_skips_sleep(self):
        with patch(
            "fault_injection.delays.time.sleep",
            side_effect=AssertionError("time.sleep should not be called when disabled"),
        ):
            self.assertIsNone(delay_inline(0.25, disable=True))


class TestDelayRandomDecorator(unittest.TestCase):
    def test_delay_random_rejects_negative_max_time(self):
        with self.assertRaisesRegex(ValueError, "delay_random should have positive max_time_s"):
            delay_random(-0.1)

    def test_delay_random_uses_randomized_delay(self):
        with patch("fault_injection.delays.random.random", return_value=0.5):
            with patch("fault_injection.delays.time.sleep") as sleep_mock:
                @delay_random(0.4)
                def add(a, b):
                    return a + b

                self.assertEqual(add(4, 6), 10)
                sleep_mock.assert_called_once_with(0.2)

    def test_delay_random_disable_skips_random_and_sleep(self):
        with patch(
            "fault_injection.delays.random.random",
            side_effect=AssertionError("random.random should not be called when disabled"),
        ):
            with patch(
                "fault_injection.delays.time.sleep",
                side_effect=AssertionError("time.sleep should not be called when disabled"),
            ):
                @delay_random(0.4, disable=True)
                def add(a, b):
                    return a + b

                self.assertEqual(add(5, 6), 11)


class TestDelayRandomInline(unittest.TestCase):
    def test_delay_random_inline_rejects_negative_max_time(self):
        with self.assertRaisesRegex(ValueError, "delay_random_inline should have positive max_time_s"):
            delay_random_inline(-0.1)

    def test_delay_random_inline_uses_randomized_delay(self):
        with patch("fault_injection.delays.random.random", return_value=0.5):
            with patch("fault_injection.delays.time.sleep") as sleep_mock:
                self.assertIsNone(delay_random_inline(0.4))
                sleep_mock.assert_called_once_with(0.2)

    def test_delay_random_inline_disable_skips_random_and_sleep(self):
        with patch(
            "fault_injection.delays.random.random",
            side_effect=AssertionError("random.random should not be called when disabled"),
        ):
            with patch(
                "fault_injection.delays.time.sleep",
                side_effect=AssertionError("time.sleep should not be called when disabled"),
            ):
                self.assertIsNone(delay_random_inline(0.4, disable=True))


class TestDelayRandomNormDecorator(unittest.TestCase):
    def test_delay_random_norm_rejects_negative_parameters(self):
        with self.assertRaisesRegex(ValueError, "delay_random_norm should have positive mean_time_s"):
            delay_random_norm(mean_time_s=-0.1, std_time_s=0.1)

        with self.assertRaisesRegex(ValueError, "delay_random_norm should have positive std_time_s"):
            delay_random_norm(mean_time_s=0.1, std_time_s=-0.1)

    def test_delay_random_norm_uses_gaussian_value_when_positive(self):
        with patch("fault_injection.delays.random.gauss", return_value=0.35):
            with patch("fault_injection.delays.time.sleep") as sleep_mock:
                @delay_random_norm(mean_time_s=0.3, std_time_s=0.1)
                def add(a, b):
                    return a + b

                self.assertEqual(add(7, 8), 15)
                sleep_mock.assert_called_once_with(0.35)

    def test_delay_random_norm_clamps_negative_gaussian_value_to_zero(self):
        with patch("fault_injection.delays.random.gauss", return_value=-0.7):
            with patch("fault_injection.delays.time.sleep") as sleep_mock:
                @delay_random_norm(mean_time_s=0.3, std_time_s=0.1)
                def add(a, b):
                    return a + b

                self.assertEqual(add(7, 8), 15)
                sleep_mock.assert_called_once_with(0)

    def test_delay_random_norm_disable_skips_gauss_and_sleep(self):
        with patch(
            "fault_injection.delays.random.gauss",
            side_effect=AssertionError("random.gauss should not be called when disabled"),
        ):
            with patch(
                "fault_injection.delays.time.sleep",
                side_effect=AssertionError("time.sleep should not be called when disabled"),
            ):
                @delay_random_norm(mean_time_s=0.3, std_time_s=0.1, disable=True)
                def add(a, b):
                    return a + b

                self.assertEqual(add(9, 1), 10)


class TestDelayRandomNormInline(unittest.TestCase):
    def test_delay_random_norm_inline_rejects_negative_parameters(self):
        with self.assertRaisesRegex(ValueError, "delay_random_norm should have positive mean_time_s"):
            delay_random_norm_inline(mean_time_s=-0.1, std_time_s=0.1)

        with self.assertRaisesRegex(ValueError, "delay_random_norm should have positive std_time_s"):
            delay_random_norm_inline(mean_time_s=0.1, std_time_s=-0.1)

    def test_delay_random_norm_inline_uses_gaussian_value_when_positive(self):
        with patch("fault_injection.delays.random.gauss", return_value=0.35):
            with patch("fault_injection.delays.time.sleep") as sleep_mock:
                self.assertIsNone(delay_random_norm_inline(mean_time_s=0.3, std_time_s=0.1))
                sleep_mock.assert_called_once_with(0.35)

    def test_delay_random_norm_inline_clamps_negative_gaussian_value_to_zero(self):
        with patch("fault_injection.delays.random.gauss", return_value=-0.7):
            with patch("fault_injection.delays.time.sleep") as sleep_mock:
                self.assertIsNone(delay_random_norm_inline(mean_time_s=0.3, std_time_s=0.1))
                sleep_mock.assert_called_once_with(0)

    def test_delay_random_norm_inline_disable_skips_gauss_and_sleep(self):
        with patch(
            "fault_injection.delays.random.gauss",
            side_effect=AssertionError("random.gauss should not be called when disabled"),
        ):
            with patch(
                "fault_injection.delays.time.sleep",
                side_effect=AssertionError("time.sleep should not be called when disabled"),
            ):
                self.assertIsNone(
                    delay_random_norm_inline(mean_time_s=0.3, std_time_s=0.1, disable=True)
                )


class TestDelayNthDecorator(unittest.TestCase):
    def setUp(self):
        if hasattr(delay_at_nth_call, "n_called_dict"):
            delay_at_nth_call.n_called_dict = {}

    def test_delay_at_n1_calls_sleep_and_returns_wrapped_value(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            @delay_at_nth_call(time_s=0.25, n=1, func_id=1)
            def add(a, b):
                return a + b

            self.assertEqual(add(1, 2), 3)
            sleep_mock.assert_called_once_with(0.25)

    def test_delay_disable_skips_sleep(self):
        with patch(
            "fault_injection.delays.time.sleep",
            side_effect=AssertionError("time.sleep should not be called when disabled"),
        ):
            @delay_at_nth_call(time_s=0.25, n=1, func_id=1, disable=True)
            def add(a, b):
                return a + b

            self.assertEqual(add(2, 3), 5)

    def test_delay_preserves_metadata_with_wraps(self):
        @delay_at_nth_call(time_s=0.25, n=1, func_id=1, disable=True)
        def add(a, b):
            return a + b

        self.assertEqual(add.__name__, "add")

    def test_delay_at_3rd_call(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            @delay_at_nth_call(time_s=0.25, n=3, func_id=1)
            def add(a, b):
                return a + b

            self.assertEqual(add(1, 2), 3)
            self.assertEqual(add(1, 2), 3)
            self.assertEqual(add(1, 2), 3)
            self.assertEqual(add(1, 2), 3)
            sleep_mock.assert_called_once_with(0.25)

    def test_explicit_func_id_uses_independent_counters_per_function(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            @delay_at_nth_call(time_s=0.4, n=2, func_id=101)
            def add(a, b):
                return a + b

            @delay_at_nth_call(time_s=0.4, n=2, func_id=102)
            def mul(a, b):
                return a * b

            self.assertEqual(add(1, 2), 3)
            self.assertEqual(mul(2, 3), 6)
            self.assertEqual(add(4, 5), 9)
            self.assertEqual(mul(3, 4), 12)
            self.assertEqual(sleep_mock.call_count, 2)

    def test_default_func_id_shares_counter_between_functions(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            @delay_at_nth_call(time_s=0.4, n=2)
            def add(a, b):
                return a + b

            @delay_at_nth_call(time_s=0.4, n=2)
            def mul(a, b):
                return a * b

            self.assertEqual(add(1, 2), 3)
            self.assertEqual(mul(2, 3), 6)
            self.assertEqual(add(4, 5), 9)
            sleep_mock.assert_called_once_with(0.4)

    def test_delay_rejects_non_positive_n(self):
        for invalid_n in (0, -1, 1.5):
            with self.subTest(invalid_n=invalid_n):
                with self.assertRaises(ValueError):
                    delay_at_nth_call(n=invalid_n)

    def test_delay_rejects_negative_time(self):
        with self.assertRaisesRegex(ValueError, "delay should have positive time_s"):
            delay_at_nth_call(time_s=-0.1)


class TestDelayNthInline(unittest.TestCase):
    def setUp(self):
        if hasattr(delay_at_nth_call_inline, "n_called_dict"):
            delay_at_nth_call_inline.n_called_dict = {}

    def test_delay_at_n1_calls_sleep(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            self.assertIsNone(delay_at_nth_call_inline(time_s=0.25, n=1, func_id=1))
            sleep_mock.assert_called_once_with(0.25)

    def test_delay_disable_skips_sleep(self):
        with patch(
            "fault_injection.delays.time.sleep",
            side_effect=AssertionError("time.sleep should not be called when disabled"),
        ):
            self.assertIsNone(delay_at_nth_call_inline(time_s=0.25, n=1, func_id=1, disable=True))

    def test_delay_at_3rd_call(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            self.assertIsNone(delay_at_nth_call_inline(time_s=0.25, n=3, func_id=1))
            self.assertIsNone(delay_at_nth_call_inline(time_s=0.25, n=3, func_id=1))
            self.assertIsNone(delay_at_nth_call_inline(time_s=0.25, n=3, func_id=1))
            self.assertIsNone(delay_at_nth_call_inline(time_s=0.25, n=3, func_id=1))
            sleep_mock.assert_called_once_with(0.25)

    def test_explicit_func_id_uses_independent_counters_per_call_site(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            def call_first():
                delay_at_nth_call_inline(time_s=0.4, n=2, func_id=1)

            def call_second():
                delay_at_nth_call_inline(time_s=0.4, n=2, func_id=2)

            self.assertIsNone(call_first())
            self.assertIsNone(call_second())
            self.assertIsNone(call_first())
            self.assertIsNone(call_second())
            self.assertEqual(sleep_mock.call_count, 2)

    def test_default_func_id_shares_counter_across_call_sites(self):
        with patch("fault_injection.delays.time.sleep") as sleep_mock:
            def call_first():
                delay_at_nth_call_inline(time_s=0.4, n=2)

            def call_second():
                delay_at_nth_call_inline(time_s=0.4, n=2)

            self.assertIsNone(call_first())
            self.assertIsNone(call_second())
            self.assertIsNone(call_first())
            sleep_mock.assert_called_once_with(0.4)

    def test_delay_inline_rejects_non_positive_n(self):
        for invalid_n in (0, -1, 1.5):
            with self.subTest(invalid_n=invalid_n):
                with self.assertRaises(ValueError):
                    delay_at_nth_call_inline(n=invalid_n)

    def test_delay_inline_rejects_negative_time(self):
        with self.assertRaisesRegex(ValueError, "delay should have positive time_s"):
            delay_at_nth_call_inline(time_s=-0.1)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import patch

from fault_injection import (
    raise_,
    raise_inline,
    raise_random,
    raise_random_inline,
    raise_at_nth_call,
    raise_at_nth_call_inline
)


class TestRaiseDecorator(unittest.TestCase):
    def test_raise_always_raises_runtime_error_when_enabled(self):
        @raise_()
        def add(a, b):
            return a + b

        with self.assertRaisesRegex(RuntimeError, "raise_ exception is raised"):
            add(1, 2)

    def test_raise_disable_calls_wrapped_function(self):
        @raise_(disable=True)
        def add(a, b):
            return a + b

        self.assertEqual(add(2, 3), 5)

    def test_raise_preserves_metadata_with_wraps(self):
        @raise_(disable=True)
        def add(a, b):
            return a + b

        self.assertEqual(add.__name__, "add")

    def test_raise_uses_custom_message(self):
        @raise_(msg="custom raise message")
        def add(a, b):
            return a + b

        with self.assertRaisesRegex(RuntimeError, "custom raise message"):
            add(1, 2)


class TestRaiseInline(unittest.TestCase):
    def test_raise_inline_raises_runtime_error_when_enabled(self):
        with self.assertRaisesRegex(RuntimeError, "raise_inline exception is raised"):
            raise_inline()

    def test_raise_inline_disable_skips_raising(self):
        self.assertIsNone(raise_inline(disable=True))

    def test_raise_inline_uses_custom_message(self):
        with self.assertRaisesRegex(RuntimeError, "inline custom message"):
            raise_inline(msg="inline custom message")


class TestRaiseRandomDecorator(unittest.TestCase):
    def test_raise_random_rejects_invalid_probability(self):
        with self.assertRaisesRegex(ValueError, "prob_of_raise should be 0-1"):
            raise_random(prob_of_raise=-0.01)

        with self.assertRaisesRegex(ValueError, "prob_of_raise should be 0-1"):
            raise_random(prob_of_raise=1.01)

    def test_raise_random_raises_when_random_value_is_below_threshold(self):
        with patch("fault_injection.raise_exception.random.random", return_value=0.19):
            @raise_random(prob_of_raise=0.2)
            def add(a, b):
                return a + b

            with self.assertRaisesRegex(RuntimeError, "raise_random exception is raised"):
                add(1, 2)

    def test_raise_random_uses_custom_message(self):
        with patch("fault_injection.raise_exception.random.random", return_value=0.1):
            @raise_random(msg="random custom message", prob_of_raise=0.2)
            def add(a, b):
                return a + b

            with self.assertRaisesRegex(RuntimeError, "random custom message"):
                add(1, 2)

    def test_raise_random_calls_wrapped_function_when_random_value_is_above_threshold(self):
        with patch("fault_injection.raise_exception.random.random", return_value=0.9):
            @raise_random(prob_of_raise=0.2)
            def add(a, b):
                return a + b

            self.assertEqual(add(3, 4), 7)

    def test_raise_random_disable_skips_random_and_raising(self):
        with patch(
            "fault_injection.raise_exception.random.random",
            side_effect=AssertionError("random.random should not be called when disabled"),
        ):
            @raise_random(prob_of_raise=1.0, disable=True)
            def add(a, b):
                return a + b

            self.assertEqual(add(10, 5), 15)


class TestRaiseRandomInline(unittest.TestCase):
    def test_raise_random_inline_rejects_invalid_probability(self):
        with self.assertRaisesRegex(ValueError, "prob_of_raise should be 0-1"):
            raise_random_inline(prob_of_raise=-0.01)

        with self.assertRaisesRegex(ValueError, "prob_of_raise should be 0-1"):
            raise_random_inline(prob_of_raise=1.01)

    def test_raise_random_inline_raises_when_random_value_is_below_threshold(self):
        with patch("fault_injection.raise_exception.random.random", return_value=0.19):
            with self.assertRaisesRegex(RuntimeError, "raise_random exception is raised"):
                raise_random_inline(prob_of_raise=0.2)

    def test_raise_random_inline_uses_custom_message(self):
        with patch("fault_injection.raise_exception.random.random", return_value=0.19):
            with self.assertRaisesRegex(RuntimeError, "inline random custom message"):
                raise_random_inline(msg="inline random custom message", prob_of_raise=0.2)

    def test_raise_random_inline_noop_when_random_value_is_above_threshold(self):
        with patch("fault_injection.raise_exception.random.random", return_value=0.9):
            self.assertIsNone(raise_random_inline(prob_of_raise=0.2))

    def test_raise_random_inline_disable_skips_random_and_raising(self):
        with patch(
            "fault_injection.raise_exception.random.random",
            side_effect=AssertionError("random.random should not be called when disabled"),
        ):
            self.assertIsNone(raise_random_inline(prob_of_raise=1.0, disable=True))


class TestRaiseNthDecorator(unittest.TestCase):
    def setUp(self):
        if hasattr(raise_at_nth_call, "n_called_dict"):
            raise_at_nth_call.n_called_dict = {}

    def test_raises_at_n1(self):
        @raise_at_nth_call(n=1, func_id=1)
        def add(a, b):
            return a + b

        with self.assertRaisesRegex(RuntimeError, "raise_at_nth_call exception is raised\nFunc id 1"):
            add(1, 2)

    def test_raise_disable_calls_wrapped_function(self):
        @raise_at_nth_call(n=1, func_id=1, disable=True)
        def add(a, b):
            return a + b

        self.assertEqual(add(2, 3), 5)

    def test_raise_preserves_metadata_with_wraps(self):
        @raise_at_nth_call(n=1, func_id=1, disable=True)
        def add(a, b):
            return a + b

        self.assertEqual(add.__name__, "add")

    def test_raise_uses_custom_message(self):
        @raise_at_nth_call(n=1, func_id=1, msg="custom raise message")
        def add(a, b):
            return a + b

        with self.assertRaisesRegex(RuntimeError, "custom raise message\nFunc id 1"):
            add(1, 2)
    
    def test_raise_at_3rd(self):
        @raise_at_nth_call(n=3, func_id=1, msg="custom raise message")
        def add(a, b):
            return a + b

        add(1, 2)
        add(1, 2)
        with self.assertRaises(RuntimeError):
            add(1, 2)

    def test_explicit_func_id_uses_independent_counters_per_function(self):
        @raise_at_nth_call(n=2, func_id=101)
        def add(a, b):
            return a + b

        @raise_at_nth_call(n=2, func_id=102)
        def mul(a, b):
            return a * b

        self.assertEqual(add(1, 2), 3)
        self.assertEqual(mul(2, 3), 6)
        with self.assertRaisesRegex(RuntimeError, r"\nFunc id 101\Z"):
            add(4, 5)
        with self.assertRaisesRegex(RuntimeError, r"\nFunc id 102\Z"):
            mul(3, 4)

    def test_default_func_id_shares_counter_between_functions(self):
        @raise_at_nth_call(n=2)
        def add(a, b):
            return a + b

        @raise_at_nth_call(n=2)
        def mul(a, b):
            return a * b

        self.assertEqual(add(1, 2), 3)
        with self.assertRaisesRegex(RuntimeError, r"\nFunc id 1\Z"):
            mul(2, 3)
        self.assertEqual(add(4, 5), 9)

    def test_raise_rejects_non_positive_n(self):
        for invalid_n in (0, -1, 1.5):
            with self.subTest(invalid_n=invalid_n):
                with self.assertRaises(ValueError):
                    raise_at_nth_call(n=invalid_n)


class TestRaiseNthInline(unittest.TestCase):
    def setUp(self):
        if hasattr(raise_at_nth_call_inline, "n_called_dict"):
            raise_at_nth_call_inline.n_called_dict = {}

    def test_raises_at_n1(self):
        with self.assertRaisesRegex(RuntimeError, "raise_at_nth_call_inline exception is raised\nFunc id 1"):
            raise_at_nth_call_inline(n=1, func_id=1)

    def test_raise_disable_skips_raising(self):
        self.assertIsNone(raise_at_nth_call_inline(n=1, func_id=1, disable=True))

    def test_raise_uses_custom_message(self):
        with self.assertRaisesRegex(RuntimeError, "inline custom nth message\nFunc id 5"):
            raise_at_nth_call_inline(n=1, func_id=5, msg="inline custom nth message")

    def test_raise_at_3rd(self):
        raise_at_nth_call_inline(n=3, func_id=1)
        raise_at_nth_call_inline(n=3, func_id=1)
        with self.assertRaisesRegex(RuntimeError, "raise_at_nth_call_inline exception is raised\nFunc id 1"):
            raise_at_nth_call_inline(n=3, func_id=1)

    def test_raise_uses_separate_counters_by_func_id(self):
        raise_at_nth_call_inline(n=2, func_id=1)
        raise_at_nth_call_inline(n=2, func_id=2)
        with self.assertRaisesRegex(RuntimeError, "Func id 1"):
            raise_at_nth_call_inline(n=2, func_id=1)
        with self.assertRaisesRegex(RuntimeError, "Func id 2"):
            raise_at_nth_call_inline(n=2, func_id=2)

    def test_explicit_func_id_uses_independent_counters_per_call_site(self):
        def call_first():
            raise_at_nth_call_inline(n=2, func_id=1)

        def call_second():
            raise_at_nth_call_inline(n=2, func_id=2)

        self.assertIsNone(call_first())
        self.assertIsNone(call_second())
        with self.assertRaises(RuntimeError):
            call_first()
        with self.assertRaises(RuntimeError):
            call_second()

    def test_default_func_id_shares_counter_across_call_sites(self):
        def call_first():
            raise_at_nth_call_inline(n=2)

        def call_second():
            raise_at_nth_call_inline(n=2)

        self.assertIsNone(call_first())
        with self.assertRaisesRegex(RuntimeError, r"\nFunc id 1\Z"):
            call_second()
        self.assertIsNone(call_first())

    def test_raise_inline_rejects_non_positive_n(self):
        for invalid_n in (0, -1, 1.5):
            with self.subTest(invalid_n=invalid_n):
                with self.assertRaises(ValueError):
                    raise_at_nth_call_inline(n=invalid_n)

if __name__ == "__main__":
    unittest.main()



"""Exception-based fault injection helpers."""

import random
from functools import wraps
from typing import Any, Callable

Decorator = Callable[[Callable[..., Any]], Callable[..., Any]]


def raise_inline(msg: str = "raise_inline exception is raised", disable: bool = False) -> None:
    """Raise ``RuntimeError`` immediately unless disabled.

    Args:
        msg: Exception message.
        disable: If ``True``, raising is skipped.
    """
    if not disable:
        raise RuntimeError(msg)


def raise_(msg: str = "raise_ exception is raised", disable: bool = False) -> Decorator:
    """Return a decorator that always raises ``RuntimeError`` unless disabled.

    Args:
        msg: Exception message. This is the first positional argument.
        disable: If ``True``, no exception is injected and the function executes.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap ``func`` with deterministic exception injection."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if disable:
                return func(*args, **kwargs)
            raise RuntimeError(msg)
        return wrapper
    return decorator


def raise_at_nth_call_inline(
        msg: str = "raise_at_nth_call_inline exception is raised",
        n: int = 5,
        func_id = 1,
        disable: bool = False
    ) -> None:
    """Raise ``RuntimeError`` at the n-th call for a given ``func_id``.

    Args:
        msg: Exception message.
        n: 1-based call number at which to raise.
        func_id: Counter key used to isolate different call sites. Calls that use the same
            ``func_id`` share the same counter.
        disable: If ``True``, raising is skipped.

    Raises:
        ValueError: If ``n`` is not a positive integer.
    """
    if n < 1 or not isinstance(n, int):
        raise ValueError("n should be a positive integer.")

    if not hasattr(raise_at_nth_call_inline, "n_called_dict"):
        raise_at_nth_call_inline.n_called_dict = {}

    if func_id not in raise_at_nth_call_inline.n_called_dict.keys():
        raise_at_nth_call_inline.n_called_dict[func_id] = 0
    raise_at_nth_call_inline.n_called_dict[func_id] += 1
    if raise_at_nth_call_inline.n_called_dict[func_id] == n and not disable:
        raise RuntimeError(msg + f"\nFunc id {func_id}")


def raise_at_nth_call(
        msg: str = "raise_at_nth_call exception is raised",
        n: int = 5,
        func_id = 1,
        disable: bool = False
    ) -> Decorator:
    """Return a decorator that raises ``RuntimeError`` on the n-th call.

    Args:
        msg: Exception message. This is the first positional argument.
        n: 1-based call number at which to raise.
        func_id: Counter key used to isolate different decorated functions. Decorators that
            use the same ``func_id`` share the same counter.
        disable: If ``True``, no exception is injected and the function executes.

    Raises:
        ValueError: If ``n`` is not a positive integer.
    """
    if n < 1 or not isinstance(n, int):
        raise ValueError("n should be a positive integer.")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap ``func`` with deterministic exception injection."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not hasattr(raise_at_nth_call, "n_called_dict"):
                raise_at_nth_call.n_called_dict = {}

            if func_id not in raise_at_nth_call.n_called_dict.keys():
                raise_at_nth_call.n_called_dict[func_id] = 0
            raise_at_nth_call.n_called_dict[func_id] += 1
            if raise_at_nth_call.n_called_dict[func_id] == n and not disable:
                raise RuntimeError(msg + f"\nFunc id {func_id}")
            return func(*args, **kwargs)
        return wrapper
    return decorator


def raise_random_inline(
    msg: str = "raise_random exception is raised",
    prob_of_raise: float = 0.1,
    disable: bool = False,
) -> None:
    """Raise ``RuntimeError`` with probability ``prob_of_raise`` unless disabled.

    Args:
        msg: Exception message.
        prob_of_raise: Probability in ``[0, 1]`` used to raise an exception.
        disable: If ``True``, raising is skipped.

    Raises:
        ValueError: If ``prob_of_raise`` is outside ``[0, 1]``.
    """
    if not 0 <= prob_of_raise <= 1:
        raise ValueError("prob_of_raise should be 0-1")
    if not disable:
        rnd = random.random()
        if rnd < prob_of_raise:
            raise RuntimeError(msg)


def raise_random(
    msg: str = "raise_random exception is raised",
    prob_of_raise: float = 0.1,
    disable: bool = False,
) -> Decorator:
    """Return a decorator that raises ``RuntimeError`` with a set probability.

    Args:
        msg: Exception message. This is the first positional argument.
        prob_of_raise: Probability in ``[0, 1]`` used to raise an exception.
        disable: If ``True``, exception injection is skipped.

    Raises:
        ValueError: If ``prob_of_raise`` is outside ``[0, 1]``.
    """
    if not 0 <= prob_of_raise <= 1:
        raise ValueError("prob_of_raise should be 0-1")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap ``func`` with probabilistic exception injection."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not disable:
                rnd = random.random()
                if rnd < prob_of_raise:
                    raise RuntimeError(msg)
            return func(*args, **kwargs)
        return wrapper
    return decorator



import unittest

from fault_injection import (
    delay,
    delay_inline,
    delay_random,
    delay_random_inline,
    delay_random_norm,
    delay_random_norm_inline,
    raise_,
    raise_inline,
    raise_random,
    raise_random_inline,
)


class TestPublicApi(unittest.TestCase):
    def test_public_api_exports_expected_symbols(self):
        self.assertTrue(callable(delay))
        self.assertTrue(callable(delay_inline))
        self.assertTrue(callable(delay_random))
        self.assertTrue(callable(delay_random_inline))
        self.assertTrue(callable(delay_random_norm))
        self.assertTrue(callable(delay_random_norm_inline))
        self.assertTrue(callable(raise_))
        self.assertTrue(callable(raise_inline))
        self.assertTrue(callable(raise_random))
        self.assertTrue(callable(raise_random_inline))


if __name__ == "__main__":
    unittest.main()



"""Delay-based fault injection helpers."""

import random
import time
from functools import wraps
from typing import Any, Callable

Decorator = Callable[[Callable[..., Any]], Callable[..., Any]]


def delay_inline(time_s: float = 0.1, disable: bool = False) -> None:
    """Inject a fixed delay immediately.

    Args:
        time_s: Sleep duration in seconds. Must be non-negative.
        disable: If ``True``, the delay is skipped.

    Raises:
        ValueError: If ``time_s`` is negative.
    """
    if time_s < 0:
        raise ValueError("delay should have positive time_s")
    if not disable:
        time.sleep(time_s)


def delay(time_s: float = 0.1, disable: bool = False) -> Decorator:
    """Return a decorator that injects a fixed delay before function execution.

    Args:
        time_s: Sleep duration in seconds. Must be non-negative.
        disable: If ``True``, the delay is skipped.

    Raises:
        ValueError: If ``time_s`` is negative.
    """
    if time_s < 0:
        raise ValueError("delay should have positive time_s")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap ``func`` with a fixed pre-execution delay."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not disable:
                time.sleep(time_s)
            return func(*args, **kwargs)
        return wrapper

    return decorator


def delay_at_nth_call_inline(
        time_s: float = 0.1,
        n: int = 5,
        func_id = 1,
        disable: bool = False
    ) -> None:
    """Inject a fixed delay at the n-th call for a given ``func_id``.

    Args:
        time_s: Sleep duration in seconds. Must be non-negative.
        n: 1-based call number at which to inject the delay.
        func_id: Counter key used to isolate different call sites. Calls that use the same
            ``func_id`` share the same counter.
        disable: If ``True``, delay is skipped.

    Raises:
        ValueError: If ``time_s`` is negative.
        ValueError: If ``n`` is not a positive integer.
    """
    if time_s < 0:
        raise ValueError("delay should have positive time_s")
    if n < 1 or not isinstance(n, int):
        raise ValueError("n should be a positive integer.")

    if not hasattr(delay_at_nth_call_inline, "n_called_dict"):
        delay_at_nth_call_inline.n_called_dict = {}

    if func_id not in delay_at_nth_call_inline.n_called_dict.keys():
        delay_at_nth_call_inline.n_called_dict[func_id] = 0
    delay_at_nth_call_inline.n_called_dict[func_id] += 1
    if delay_at_nth_call_inline.n_called_dict[func_id] == n and not disable:
        time.sleep(time_s)


def delay_at_nth_call(
        time_s: float = 0.1,
        n: int = 5,
        func_id = 1,
        disable: bool = False
    ) -> Decorator:
    """Return a decorator that injects a fixed delay on the n-th call.

    Args:
        time_s: Sleep duration in seconds. Must be non-negative.
        n: 1-based call number at which to inject the delay.
        func_id: Counter key used to isolate different decorated functions. Decorators that
            use the same ``func_id`` share the same counter.
        disable: If ``True``, delay is skipped.

    Raises:
        ValueError: If ``time_s`` is negative.
        ValueError: If ``n`` is not a positive integer.
    """
    if time_s < 0:
        raise ValueError("delay should have positive time_s")
    if n < 1 or not isinstance(n, int):
        raise ValueError("n should be a positive integer.")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not hasattr(delay_at_nth_call, "n_called_dict"):
                delay_at_nth_call.n_called_dict = {}
            if func_id not in delay_at_nth_call.n_called_dict.keys():
                delay_at_nth_call.n_called_dict[func_id] = 0
            delay_at_nth_call.n_called_dict[func_id] += 1
            if delay_at_nth_call.n_called_dict[func_id] == n and not disable:
                time.sleep(time_s)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def delay_random_inline(max_time_s: float = 0.1, disable: bool = False) -> None:
    """Inject a uniform random delay immediately.

    Args:
        max_time_s: Maximum sleep duration in seconds. Must be non-negative.
        disable: If ``True``, the random delay is skipped.

    Raises:
        ValueError: If ``max_time_s`` is negative.
    """
    if max_time_s < 0:
        raise ValueError("delay_random_inline should have positive max_time_s")
    if not disable:
        rnd = random.random()
        time_s = max_time_s * rnd
        time.sleep(time_s)


def delay_random(max_time_s: float = 0.1, disable: bool = False) -> Decorator:
    """Return a decorator that injects a uniform random delay before execution.

    Args:
        max_time_s: Maximum sleep duration in seconds. Must be non-negative.
        disable: If ``True``, the random delay is skipped.

    Raises:
        ValueError: If ``max_time_s`` is negative.
    """
    if max_time_s < 0:
        raise ValueError("delay_random should have positive max_time_s")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap ``func`` with a random delay in ``[0, max_time_s]``."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not disable:
                rnd = random.random()
                time_s = max_time_s * rnd
                time.sleep(time_s)
            return func(*args, **kwargs)
        return wrapper

    return decorator


def delay_random_norm_inline(
    mean_time_s: float = 0.3,
    std_time_s: float = 0.1,
    disable: bool = False,
) -> None:
    """Inject a Gaussian random delay immediately.

    The sampled delay is clamped to zero to avoid negative sleep times.

    Args:
        mean_time_s: Mean of the Gaussian distribution in seconds.
        std_time_s: Standard deviation of the Gaussian distribution in seconds.
        disable: If ``True``, the random delay is skipped.

    Raises:
        ValueError: If ``mean_time_s`` or ``std_time_s`` is negative.
    """
    if mean_time_s < 0:
        raise ValueError("delay_random_norm should have positive mean_time_s")
    if std_time_s < 0:
        raise ValueError("delay_random_norm should have positive std_time_s")
    if not disable:
        time_s = random.gauss(mean_time_s, std_time_s)
        time_s = max(0, time_s)
        time.sleep(time_s)


def delay_random_norm(
    mean_time_s: float = 0.3,
    std_time_s: float = 0.1,
    disable: bool = False,
) -> Decorator:
    """Return a decorator that injects a Gaussian random delay before execution.

    The sampled delay is clamped to zero to avoid negative sleep times.

    Args:
        mean_time_s: Mean of the Gaussian distribution in seconds.
        std_time_s: Standard deviation of the Gaussian distribution in seconds.
        disable: If ``True``, the random delay is skipped.

    Raises:
        ValueError: If ``mean_time_s`` or ``std_time_s`` is negative.
    """
    if mean_time_s < 0:
        raise ValueError("delay_random_norm should have positive mean_time_s")
    if std_time_s < 0:
        raise ValueError("delay_random_norm should have positive std_time_s")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Wrap ``func`` with a non-negative Gaussian random delay."""
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            if not disable:
                time_s = random.gauss(mean_time_s, std_time_s)
                time_s = max(0, time_s)
                time.sleep(time_s)
            return func(*args, **kwargs)
        return wrapper

    return decorator



from .delays import (delay_inline, delay, delay_random_inline, delay_random,
    delay_random_norm_inline, delay_random_norm, delay_at_nth_call_inline, delay_at_nth_call)
from .raise_exception import (raise_inline, raise_, raise_at_nth_call,
    raise_at_nth_call_inline, raise_random_inline, raise_random)
